/*
Specshow Color Spectra Visualization Program
============================================

This program package consists of the following files.

Makefile
README
asc2spes.c
initfns.c
specshow.c
spekutil.c
spekutils.h

%************************************************************************
% Specshow Color Spectra Visualization Program			        *
% ============================================			        *
%                                                                       *
%  (C) Ari Saarinen 1993-1994						*
%  Lappeenranta University of Technology, Finland 			*
%                                                                       *
%  NOTE: This program is copyrighted in the sense that it               *
%  may be used for scientific purposes. The program as a whole, or      *
%  parts thereof, cannot be included or used in any commercial          *
%  application without written permission granted by its producents.    *
%  This program comes 'as it is' and with no warranty.                  *
%                                                                       *
%  All comments concerning this program may be sent to the              *
%  e-mail address 'haanpalo@lut.fi'.                                    *
%                                                                       *
%***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>

#include <spekutils.h>

int printout = 0; /* print error messages etc. when testing */

int comb_sps(), read2mem(), read_speksize(); 
int count_RGB_values(), calc_values(), make_spek(), check_uniform();
int mul_spek(), copy_spec(), copy_values();

double count_value();


/* 
  comb_sps()
  purpose: multiply two spectrums sp1 and sp2 with each other;
  sp1 and sp2 must be have same start and end-wavelens and same
  size. Scales new values betw. 0 and 1. Returns 0 if not succes, 
  1 otherwise
*/
int comb_sps(sp1, sp2)
	Spektri *sp1, *sp2;
{
	int i;
	char strb[200];

	/* check the uniformity of two spectrums */
	if(sp1->alku != sp2->alku || sp1->loppu != sp2->loppu
		|| sp1->lkm != sp2->lkm)
		return 0;

	for(i = 0; i < sp1->lkm; i++){
		sp1->spek[i] = (sp1->spek[i]/sp1->spek_max) * (sp2->spek[i] / sp2->spek_max);
	}
	sp1->spek_min = 0.0;
	sp1->spek_max = 1.0;
	sp1->type = NORMAL;
	sprintf(strb, "%s + %s", sp1->comment, sp2->comment);
	if(strlen(strb) > COM_MAX_SIZE)
		strb[COM_MAX_SIZE - 1] = '\0';
	strcpy(sp1->comment, strb);
	return 1;
}

/*
  	count_rgb()
 	purpose: Count r, g and b values of spectrum sp in coord. system coord.
  	Multiplies spectrum with 'ke', if wanted lightened rgb-values. Returns 
  	0 if not success.
*/
int count_rgb(r, g, b, sp, ke, coord)
     	double *r, *g, *b;
	int ke;
     	Spektri *sp;
     	Coord* coord;
{
     	double k, X, Y, Z;
     	double *xline, *yline, *zline;

        xline = coord->xline.spek;
     	yline = coord->yline.spek;
     	zline = coord->zline.spek;

	WHITE_SPEK.lkm = sp->lkm; 
	WHITE_SPEK.alku = sp->alku;
	WHITE_SPEK.loppu = sp->loppu;

	if(sp->type == WHITE) 
		sp->lkm = coord->xline.lkm;

	k = 100.0 / (count_value(&WHITE_SPEK, &(coord->yline), 1));

	X = k * count_value(sp, &(coord->xline), ke);
	Y = k * count_value(sp, &(coord->yline), ke);
	Z = k * count_value(sp, &(coord->zline), ke);

	*r = 1.91*X - 0.532*Y - 0.288*Z;
	*g = -0.985*X + 2.0*Y - 0.028*Z;
	*b = 0.058*X - 0.118*Y + 0.898*Z;
}

/*
  	calc_values()
  	purpose: count values of typedef type Values from spectrum 'spek'.
  	Uses coord. system 'coord'. Returns 0 if not success.
*/
int calc_values(values, spek, coord)
	Values *values;
	Spektri *spek;
	Coord *coord;
{
	double k, X, Y, Z, X0, Y0, Z0, x, y, z;
	double nim, up, vp, up0, vp0;
	double *xviiva, *yviiva, *zviiva;
	int i;

	/* set the form of WHITE_SPEK to same as spek */
	WHITE_SPEK.alku = spek->alku;
	WHITE_SPEK.loppu = spek->loppu;
	WHITE_SPEK.lkm = spek->lkm;
	if(spek->type == WHITE) 
		spek->lkm = coord->xline.lkm;

	k = 100.0 / (count_value(&WHITE_SPEK, &(coord->yline), 1));
	X = k * count_value(spek, &(coord->xline), 1);
	Y = k * count_value(spek, &(coord->yline), 1);
	Z = k * count_value(spek, &(coord->zline), 1);

	nim = X + Y + Z;
	values->Y = Y;
	values->Z = Z;
	values->X = X;	
	
	values->x = x = X / nim; 
	values->y = y = Y / nim;
	values->z = z = Z / nim;

	values->R = 1.91*X - 0.532*Y - 0.288*Z;
	values->G = -0.985*X + 2.0*Y - 0.028*Z;
	values->B = 0.058*X - 0.118*Y + 0.898*Z;

	if(x <0 || y<0 || z<0 || Y < 0){
		printf("x,y,z or Y < 0\n");
		return 0;
	}

	values->u = (4 * x) / (-2*x + 12* y + 3);
	values->v = (6 * y) / (-2*x + 12*y + 3);

	if(values->u <0 || values->v <0){
		printf("u or v < 0\n");
		return 0;
	}

	X0 = k * count_value(&WHITE_SPEK, &(coord->xline), 1);
	Y0 = k * count_value(&WHITE_SPEK, &(coord->yline), 1);
	Z0 = k * count_value(&WHITE_SPEK, &(coord->zline), 1);

	values->Lt = 116.0 * pow((Y/Y0), (1.0/3.0)) - 16.0;
	
	up = 4*X / (X + 15*Y + 3*Z);
	vp = 9*Y / (X + 15*Y + 3*Z);
	up0 = 4*X0 / (X0 + 15*Y0 + 3*Z0);
	vp0 = 9*Y0 / (X0 + 15*Y0 + 3*Z0),

	values->ut = 13 * values->Lt * (up - up0);
	values->vt = 13 * values->Lt * (vp - vp0);

	values->at = 500 * (pow((X/X0), (1.0/3.0)) - pow((Y/Y0), (1.0/3.0)));
	values->bt = 200 * (pow((Y/Y0), (1.0/3.0)) - pow((Z/Z0), (1.0/3.0)));
	
	if(printout){
		printf("name: %s\n", spek->comment);
		
		printf("x = %3.2lf, y = %3.2lf, Y = %lf\n", values->x, values->y, values->Y);
		printf("L* = %lf, a* = %lf, b* = %lf\n", values->Lt, values->at, values->bt);
		printf("u* = %lf, v* = %lf\n", values->ut, values->vt);
	}

	return 1;
}


/*
  	check_uniform()
  	purpose: check if spek1 and spek2 has the same form (size etc.) 
  	if so, return non-zero, otherwise return 0.
*/
int check_uniform(spek1, spek2)
	Spektri *spek1;
	Spektri *spek2;
{
	if(spek1->alku != spek2->alku ||
	spek1->loppu != spek2->loppu ||
	spek1->lkm != spek2->lkm)
		return 0;
	
	return 1;
}
	

/*
  	count_value()
 	count the sum of product of parameters sp and viiva. Multiply sp 
 	with k (for lightening spectrums, normally k=1). Check if type is
 	WHITE, if so, calculation is simplified. Return sum.
*/
double count_value(sp, viiva, k)
	Spektri *sp;
	Spektri *viiva;
	int k;
{
	int i1=0, i2=0, w, s1, s2, wstep, add1, add2;
	double summa = 0, kerr;

	s1 = (sp->loppu - sp->alku)/(sp->lkm -1);
	s2 = (viiva->loppu -viiva->alku)/(viiva->lkm -1);
	add1 = MAX(s1, s2)/s1;
	add2 = MAX(s1, s2)/s2;

	wstep = MAX(s1, s2);

	for(w = MIN(sp->alku, viiva->alku); w <= sp->loppu  && w <= viiva->loppu;
	w += wstep){
		if(w < sp->alku)
			kerr = 0.0;
		else if(sp->type == WHITE){
			kerr = 1.0;
			i1 += add1;
		}
		else{
			kerr = k* sp->spek[i1]/sp->spek_max;
			i1 += add1;
		}
		if(w >= viiva->alku){
			summa += kerr * viiva->spek[i2];
			i2 += add2;
        	}
	}

	return summa;
}


/* 
  	mul_spek()
  	purpose: multiply all data-values of spectrum 'sp' with coefficient
  	k. If result is greater than spec_max of sp-structure, set the value
  	to sp->spek_max
*/
int mul_spek(sp, k)
	Spektri *sp;
	int k;
{
  	int i;

  	for(i = 0; i<sp->lkm; i++){
    		sp->spek[i] *= k;
    		if(sp->spek[i] > sp->spek_max)
      			sp->spek[i] = sp->spek_max;
 	}
}


/*
  	print_sp()
  	purpose: prints the header (and data-values of 'sp' if table !=0)
  	Mainly for test purposes (not used in present programs), but may 
 	be useful later, so included here 
*/
print_sp(sp, table)
	Spektri *sp;
	int table; /* if not zero, print table (values) too */
{
	int i;

	printf("alku = %d, loppu = %d\n", sp->alku, sp->loppu);
	printf("lkm = %d\n", sp->lkm);
	printf("min = %lf, max = %lf\n", sp->spek_min, sp->spek_max);
	if(table){
		for(i=0; i< sp->lkm; i++)
			printf("%lf\t", sp->spek[i]);
	}
}


/*
  	read_speksize()
  	purpose: read the size and type of spectrum file 'fname'. Return 0
  	if everything ok. Place size and type to parameters
*/
int read_speksize(fname, type, size)
	char *fname;
	int *type;
	int *size;
{
	FILE *fp;
	int alku, loppu, koko, tyyppi;

	if((fp = fopen(fname,"r"))==NULL)
		return FILE_NOT_FOUND;
	
	if(fscanf(fp, "%d%d%d%d", &alku, &loppu, &koko, &tyyppi) != 4)
		return BAD_HEADER;

	if(alku == 0 && loppu == 0)
		return BAD_HEADER;

	fclose(fp);
	*type = tyyppi;
	*size = koko;
	return 0;
}

/* 
  	copy_spec()
  	purpose: copy Spektri-structure src to dest. Notice that space for
  	dest must be allocated before calling this function
*/
int copy_spec(dest, src)
	Spektri *dest, *src;
{
	int i;
	dest->alku = src->alku;
	dest->loppu = src->loppu;
	dest->lkm = src->lkm;
	dest->type = src->type;
	dest->spek_min = src->spek_min;
	dest->spek_max = src->spek_max;
	strcpy(dest->comment, src->comment);
	for(i=0; i<src->lkm; i++)
		dest->spek[i] = src->spek[i];
	return 1;
}


/* 
  	copy_values()
  	purpose: copy Values-structure src to dest. Notice that space for
  	dest must be allocated before calling this function
*/
int copy_values(dest, src)
	Values *dest, *src;
{
	dest->x = src->x; dest->y = src->y; dest->z = src->z;
	dest->u = src->u; dest->v = src->v;
	dest->X = src->X; dest->Y = src->Y; dest->Z = src->Z;
	
	dest->R = src->R; dest->G = src->G; dest->B = src->B; 
	
	dest->Lt = src->Lt; dest->ut = src->ut; dest->vt = src->vt;
	dest->at = src->at; dest->bt = src->bt;
	return 1;
}


/*
  	read2mem()
  	purpose: read spectrumfile 'fname' into Spektri-structure pointed
  	by 'spektri'. Space for structure must be allocated before calling
  	this function (see make_spek()-function). Data in file may be either
  	doubles or integers. Returns 0 if ok.
*/
read2mem(fname, spektri)
	char *fname;
	Spektri *spektri;
{
	int lkm, alku, loppu, type;
	int *tempi; double *tempd;
	double min, max;
	char combuf[400];
	FILE *fp;
	char ch[200];
	int len, klask, fret, i;

	if((fp = fopen(fname,"rb"))==NULL)
		return FILE_NOT_FOUND;

	if(fscanf(fp, "%d%d%d%d%lf%lf\n", &alku, &loppu, &lkm, &type, &min, &max) != 6){
		if(printout)
			printf("bad header in read2mem()");
		return BAD_HEADER;
	}

	/* check if comment is used (comment starts with
		"*"-character)
	*/
	fgets(ch, 199, fp);
	if(ch[0] != '*')
		return COMMENT_BEG_MARK_NOT_FOUND;
	len = 0; klask = 0;

	while(1){
		fgets(ch, 199, fp);
		if(ch[0] =='*') 
			break;
		ch[strlen(ch)-1]='\0';
		strcpy(&combuf[len], ch);
		len += strlen(ch); klask ++;	
		/* if comment length is more than 5 lines, something's 
		   wrong in file */
		if(klask >5) 
			return COMMENT_ENDMARK_NOT_FOUND;
	}

	/* memory allocation done before calling this function */
	spektri->alku = alku;
	spektri->loppu = loppu;
	spektri->lkm = lkm;
	spektri->type = type;
	spektri->spek_min = min;
	spektri->spek_max = max;

	/* malloc local space for data and read data into it */
	if(type == INTE){
		if((tempi = (int *)malloc((lkm) * sizeof(int))) == NULL){
			printf("malloc failed in read2mem();\n");
			exit(1);
		}

		if((fret = fread(tempi, sizeof(int), lkm, fp)) != lkm){
			if(printout)
				printf("int fread returned %d\n", fret);
			return FILE_SIZE_NOT_OK;
		}

		for(i = 0; i<lkm; i++)
			spektri->spek[i] = tempi[i];
		free(tempi);
	}
	else{
		if((tempd = (double *)malloc(lkm * sizeof(double))) == NULL){
			printf("malloc failed in read2mem();\n");
			exit(1);
		}

		if((fret = fread(tempd, sizeof(double), lkm, fp)) != lkm){
			if(printout)
				printf("double fread returned %d\n", fret);
			return FILE_SIZE_NOT_OK;
		}

		for(i =0; i<lkm; i++)
			spektri->spek[i] = tempd[i];
		free(tempd);
	}


	if(strlen(combuf)> COM_MAX_SIZE)
		combuf[COM_MAX_SIZE-1] = '\0';
		
	strcpy(spektri->comment, combuf);

	/* ok */
	fclose(fp);
	return 0;
}


/* 	
	make_spek()
	purpose: read spectrum file 'name' into Spektri-strcuture 'spek'.
	Read size of file using read_speksize(), allocs memory for spek
	and read then data into memory using read2mem(). If making 
	a monochromatic spectrum, type must be set to MONO before
	calling this function. Returns 0 if fails, 1 otherwise
*/
int make_spek(name, spek)
	char *name;
	Spektri *spek;
{
	int ind, i, mlkm, step;
	int size, type, errcode;
	char errbuf[100];

	/* if type is MONO, make a monochromatic spectrum */
	if(spek->type == MONO){
		if((ind = atoi(&name[5]))==0){
			return 0;
		}
		mlkm = spek->lkm;
		step = (spek->loppu - spek->alku)/(spek->lkm -1);
		ind = (ind - spek->alku) / step;
		if(ind <0 || ind > 61){
			return 0;
		}

		if((spek->spek = (double*)malloc(sizeof(double)*mlkm)) == NULL){
			fprintf(stderr,"make_spek():malloc failed\n");
			exit(1);
		}
	
		/* fill header, other header info already in spek */
		spek->type = NORMAL;
		spek->spek_max = 1.0;
		spek->spek_min = 0.0;

		/* set the appropriate index to 1, others equal 0 */
		for(i=0; i<spek->lkm; i++)
			spek->spek[i] = 0.0;
		spek->spek[ind]=1;
		strcpy(spek->comment, "monochromatic");
	}

	/* if type is WHITE, make a white spectrum (all values=1)
	   NOT USED */
	else if(spek->type == WHITE){
		if((spek->spek = (double*)malloc(sizeof(double)*61)) == NULL){
			fprintf(stderr,"make_spek():malloc failed\n");
			exit(1);
		}
		spek->alku = 400;
		spek->loppu = 700;
		spek->lkm = 61;
		spek->spek_max = 1.0;
		spek->spek_min = 0.0;
		for(i=0; i<spek->lkm; i++)
			spek->spek[i] = 1.0;

		strcpy(spek->comment, "white spectrum");
	}
	/* else type is normal */
	else {
		if((errcode = read_speksize(name, &type, &size)) != 0){
			switch(errcode){
			case FILE_NOT_FOUND:
				if(printout)
					printf("couldn't open file %s\n", name);
				break;
			case BAD_HEADER:
				if(printout)
					printf("bad header in file\n", name);
				break;
			default: /* should not happen */
				if(printout)
					printf("header read ret. not zero\n");
				break;
			}
			return 0;
		}
		/* alloc space for spek */
		if((spek->spek = (double*)malloc(sizeof(double)
			*(size)))==NULL){
			fprintf(stderr,"make_spek(): malloc failed\n");
			exit(1);
		}
	      	
		/* read file to memory */
		if((errcode = read2mem(name, spek)) != 0){
			printf("read2mem ret. not zero");
			return 0;
		}		
	}
	
	/* everything is ok */
	return 1;
}















